
/***********************************************************************\
*                                                                       *
*  Module      : fifoctlr_ccmw2_v2.v            Last Update: 08/08/02   *
*                                                                       *
*  Description : FIFO controller top level.                             *
*                Implements a 511x36 Read, 2044x9 Write FIFO with       *
*                common read/write clocks.                              *
*                                                                       *
*  The following Verilog code implements a 511x36 Read, 2044x9 Write    *
*  FIFO in a Virtex2 device.  The inputs are a Clock, a Read Enable, a  *
*  Write Enable, Write Data, and a FIFO_gsr signal as an initial reset. *
*  The outputs are Read Data, Full, Empty, and the FIFOcount outputs,   *
*  which indicate how full the FIFO is.                                 *
*                                                                       *
*  Designer    : Nick Camilleri                                         *
*                                                                       *
*  Company     : Xilinx, Inc.                                           *
*                                                                       *
*  Disclaimer  : THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY    *
*                WHATSOEVER AND XILINX SPECIFICALLY DISCLAIMS ANY       *
*                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR     *
*                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.         *
*                THEY ARE ONLY INTENDED TO BE USED BY XILINX            *
*                CUSTOMERS, AND WITHIN XILINX DEVICES.                  *
*                                                                       *
*                Copyright (c) 2002 Xilinx, Inc.                        *
*                All rights reserved                                    *
*                                                                       *
\***********************************************************************/

`timescale 1ns / 10ps

`define DATA_WIDTHR 35:0
`define DATA_WIDTHW 8:0
`define ADDR_WIDTH 8:0
`define MINORADDR_WIDTH 1:0

module fifoctlr_ccmw2_v2 (clock_in, read_enable_in, write_enable_in, 
                          write_data_in, fifo_gsr_in, read_data_out, 
                          full_out, empty_out, fifocount_out );

input clock_in, read_enable_in, write_enable_in, fifo_gsr_in;
input  [`DATA_WIDTHW] write_data_in;
output [`DATA_WIDTHR] read_data_out;
output full_out, empty_out;
output [3:0] fifocount_out;

wire read_enable = read_enable_in;
wire write_enable = write_enable_in;
wire fifo_gsr = fifo_gsr_in;
wire [`DATA_WIDTHW] write_data = write_data_in;
wire [`DATA_WIDTHR] read_data;
assign read_data_out = read_data;
reg full, empty;
assign full_out = full;
assign empty_out = empty;

reg [`ADDR_WIDTH] read_addr, write_addr, fcounter;
reg [`MINORADDR_WIDTH] write_addr_minor;
reg last_subword;

wire read_allow, write_allow, write_allow_minor, fcnt_allow;

wire [`DATA_WIDTHR] gnd_bus = 'h0;
wire gnd = 0;
wire pwr = 1;

/**********************************************************************\
*                                                                      *
*  A global buffer is instantianted to avoid skew problems.            * 
*                                                                      *
\**********************************************************************/
 
BUFGP gclk1 (.I(clock_in), .O(clock));

/**********************************************************************\
*                                                                      *
*  Block Select RAM instantiation for FIFO.  Module is 512x36, of      * 
*  which one address location is sacrificed for the overall speed of   *
*  the design.  NOTE: the 36-bit read data will come out LSB to MSB    *
*  relative to the original 9-bit words written.  e.g. the first 9-bit *
*  word written will be readword[8:0], then [17:9], [26:18] and        *
*  finally [35:27].                                                    *
*                                                                      *
\**********************************************************************/

RAMB16_S9_S36 bram1 ( .ADDRB(read_addr), .ADDRA({write_addr, write_addr_minor}),
                      .DIB(gnd_bus[35:4]), .DIPB(gnd_bus[3:0]), 
                      .DIA(write_data[8:1]), .DIPA(write_data[0:0]), .WEB(gnd), 
                      .WEA(pwr), .CLKB(clock), .CLKA(clock), .SSRB(gnd), 
                      .SSRA(gnd), .ENB(read_allow), .ENA(write_allow_minor), 
                      .DOB({read_data[35:28], read_data[26:19], read_data[17:10],
                            read_data[8:1]}), .DOPB({read_data[27], read_data[18],
                            read_data[9], read_data[0]}), .DOA(), .DOPA() );

/***********************************************************\
*                                                           *
*  Set allow flags, which control the clock enables for     *
*  read, write, and count operations.                       *
*                                                           *
\***********************************************************/

assign read_allow = read_enable && ! empty;
assign write_allow = (write_enable && ! full && last_subword);
assign write_allow_minor = write_enable && ! full;
assign fcnt_allow = write_allow ^ read_allow;

/***********************************************************\
*                                                           *
*  Last_subword is set when the minor address (either read  *
*  or write, whichever the design uses) is all 1's.  This   *
*  allows a slightly faster implementation.                 *
*                                                           *
\***********************************************************/

always @(posedge clock or posedge fifo_gsr)
   if (fifo_gsr) last_subword <= 0;
   else if (write_allow_minor) last_subword <= (write_addr_minor == 2'b10);

/***********************************************************\
*                                                           *
*  Empty flag is set on fifo_gsr (initial), or when on the  *
*  next clock cycle, Write Enable is low, and either the    *
*  FIFOcount is equal to 0, or it is equal to 1 and Read    *
*  Enable is high (about to go Empty).                      *
*                                                           *
\***********************************************************/

wire [3:0] fcntandout;
wire ra_or_fcnt0, wa_or_fcnt0, emptyg, fullg;

assign ra_or_fcnt0 = ((read_enable && ! empty) || ! fcounter[0]);

fifoand4b4 emptyand1 (.data(fcounter[4:1]), .out(fcntandout[0]));
fifoand4b4 emptyand2 (.data(fcounter[8:5]), .out(fcntandout[1]));
fifoand4b1 emptyand3 (.in1(fcntandout[0]), .in2(fcntandout[1]), .in3(ra_or_fcnt0),
                      .in4(write_allow), .out(emptyg));
 
always @(posedge clock or posedge fifo_gsr)
   if (fifo_gsr) empty <= 1;
   else empty <= emptyg;

/***********************************************************\
*                                                           *
*  Full flag is set on fifo_gsr (but it is cleared on the   *
*  first valid clock edge after fifo_gsr is removed), or    *
*  when on the next clock cycle, Read Enable is low, and    *
*  either the FIFOcount is equal to FF (hex), or it is      *
*  equal to FE and the Write Enable is high (about to go    *
*  Full).                                                   *
*                                                           *
\***********************************************************/

assign wa_or_fcnt0 = ((write_enable && ! full && last_subword) || fcounter[0]);

fifoand4   fulland1 (.data(fcounter[4:1]), .out(fcntandout[2]));
fifoand4   fulland2 (.data(fcounter[8:5]), .out(fcntandout[3]));
fifoand4b1 fulland3 (.in1(fcntandout[2]), .in2(fcntandout[3]), .in3(wa_or_fcnt0),
                     .in4(read_allow), .out(fullg));

always @(posedge clock or posedge fifo_gsr)
   if (fifo_gsr) full <= 1;
   else full <= fullg;

/************************************************************\
*                                                            *
*  Generation of Read and Write address pointers.  They now  *
*  use binary counters because it is simpler in simulation,  *
*  and the previous LFSR implementation wasn't in the        *   
*  critical path.                                            *
*                                                            *
\************************************************************/

always @(posedge clock or posedge fifo_gsr)
   if (fifo_gsr) read_addr <= 'h0;
   else if (read_allow)
      read_addr <= read_addr + 1;

always @(posedge clock or posedge fifo_gsr)
   if (fifo_gsr) write_addr <= 'h0;
   else if (write_allow)
      write_addr <= write_addr + 1;

/************************************************************\
*                                                            *
*  Generation of minor addresses for read operations.  Bits  *
*  are sent to the LSB of the Block SelectRAM write address, *
*  and are used to determine when to increment write_addr.   *
*                                                            *
\************************************************************/

always @(posedge clock or posedge fifo_gsr)
   if (fifo_gsr) write_addr_minor <= 'h0;
   else if (write_allow_minor)
      write_addr_minor <= write_addr_minor + 1;
 
/************************************************************\
*                                                            *
*  Generation of FIFOcount outputs.  Used to determine how   *
*  full FIFO is, based on a counter that keeps track of how  *
*  many words are in the FIFO.  Also used to generate Full   *
*  and Empty flags.  Only the upper four bits of the counter *
*  are sent outside the FIFO module.                         *
*                                                            *
\************************************************************/

always @(posedge clock or posedge fifo_gsr)
   if (fifo_gsr) fcounter <= 'h0;
   else if (fcnt_allow)
      fcounter <= fcounter + { read_enable && ! empty, read_enable && ! empty, 
                               read_enable && ! empty, read_enable && ! empty, 
                               read_enable && ! empty, read_enable && ! empty, 
                               read_enable && ! empty, read_enable && ! empty, 
                               read_enable && ! empty, pwr };

assign fifocount_out = fcounter[8:5];

endmodule

/************************************************************\
*                                                            *
*  The logic modules below are coded explicitly, to ensure   *
*  that the logic is implemented in a minimum of levels.     *   
*                                                            *
\************************************************************/

module fifoand4 (data, out);

input [3:0] data;
output out;

assign #1 out = (data[3] && data[2] && data[1] && data[0]);

endmodule

module fifoand4b4 (data, out);

input [3:0] data;
output out;

assign #1 out = (! data[3] && ! data[2] && ! data[1] && ! data[0]);

endmodule

module fifoand4b1 (in1, in2, in3, in4, out);

input in1, in2, in3, in4;
output out;

assign #1 out = (in1 && in2 && in3 && ! in4);

endmodule

